/*
 * Support for IBM PPC 405EP processors.
 *
 * Author: Charles Lee
 *         WELCM <charles@welcm.com>
 *
 * 2007 (c) WELCM.  This file is licensed under the terms of 
 * the GNU General Public License. This program is licensed "as is"
 * without any warranty of any kind, whether express or implied.
 */

/*
 * This module handles the interrupt from EBC. The interrupt is connected
 * to 1 PPS signal. Events are generated for PPS driver. If your kernel does not have
 * PPS capability, undefined symbols may occur when this module is loaded.
 *
 * Because PPS signal is edge trigger, changes are made to
 * struct ppc4xx_uic_settings ppc4xx_core_uic_cfg[] __initdata in 
 * ppc/platform/4xx/ibm405ep.c:
 *					  Default		PPS			(Caution, see below)
 *	{ .polarity 	= 0xffff7f80,	0xffff7f82,	(0xffff7f82,)
 *	  .triggering	= 0x00000000,	0x00000002,	(0x00000000,)
 *	  .ext_irq_mask	= 0x0000007f,	0x0000007d,	(0x0000007d,)
 *	}
 * The default is negative polarity and level trigger. The EBC pin 36, interrupt pin, is
 * pull-up to Vcc. Thus, the default to negative polarity is safe not to generate interrupt.
 * In addition, polarity and triggering settings are ignored if the mask bit is 1.
 *
 * Caution (see the settings in the caution colum above):
 * If you use position polarity and level triggering and leave pin 36 open, the system
 * will not response to the console. The system "hangs" because interrupts are constantly
 * occuring. Shorting the pin 36 to pin 37 or 38 (both are ground) will bring the console
 * back.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/workqueue.h>	// some functions are EXPORT_SYMBOL_GPL(),
								// such as __create_workqueue() and queue_work()
#include <linux/interrupt.h>

#include <asm/io.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <linux/delay.h>

#include <linux/timepps.h>

// always enable PPS debug
#ifndef CONFIG_PPS_DEBUG
#define CONFIG_PPS_DEBUG
#endif

#include <linux/pps.h>

// Sample from The Linux Kernel Module Programming Guides uses work queue.
// Hoever, some of work queue related functions are EXPORT_SYMBOL_GPL().
// Thus if MODULE_LICENSE() is not GPL, undedfined symbols will be at runtime

// Tasklet from Linux Device Drivers, 2nd Edition works as well. No GPL required
// However, a taints kernel message will be shown for Proprietary license

// work queue related codes are commented out. They works with GPL.
//#define MY_WORK_QUEUE_NAME "WQsched.c"

//static struct workqueue_struct *ebcWorkQueue;

// following struct is for passing info from 1st half to 2nd half
//struct ebcIntrInfo {
//	int info;
//};

/*
 * This is the "bottom" half of interrupt handling.
 */ 
//static void do2ndHalf(void *info)
//{
//	// simulate the end of 1 PPS to clear the event
//	linuxpps_event(1, PPS_CAPTURECLEAR);
//}

static void tasklet2ndHalf(unsigned long unuse)
{
	dbg("EBC intr 2nf half at %lu", jiffies);

	//	linuxpps_event(up->source, PPS_CAPTURECLEAR);
	// simulate the end of 1 PPS to clear the event
	linuxpps_event(1, PPS_CAPTURECLEAR);
}


DECLARE_TASKLET (ebcTasklet, tasklet2ndHalf, 0);

/*
 * This function handles the interrupt from EBC.
 */
static irqreturn_t ebcInterruptHandler(int irq, void *dev_id, struct pt_regs *regs)
{
/*
	static int inited = 0;
	static struct work_struct task;
	static struct ebcIntrInfo info;
*/

	dbg("EBC assert event at %lu", jiffies);

//	linuxpps_event(up->source, PPS_CAPTUREASSERT);
	linuxpps_event(1, PPS_CAPTUREASSERT);

	tasklet_schedule(&ebcTasklet);

	return IRQ_HANDLED;
}

#define IRQ_EBC 30		// From 405EP user manual table 10-1, 30 for external IRQ5

static int initEDC(void)
{
	dbg("EBC intr handler inserted at IRQ%d", IRQ_EBC);

//	ebcWorkQueue = create_workqueue(MY_WORK_QUEUE_NAME);

	return request_irq(IRQ_EBC,
					   ebcInterruptHandler,
					   SA_INTERRUPT, //SA_SHIRQ, // let's be fast
					   "EBC_Interrupt",	// name in /proc/interrupts
					   (void *) ebcInterruptHandler	// dev id, to be used in free_irq()
					   );
}

static void exitEDC(void)
{
	dbg("EBC intr handler removed");

	free_irq(IRQ_EBC, ebcInterruptHandler);
}

module_init(initEDC);
module_exit(exitEDC);

// Specify licence, author, description, and suppoted device
MODULE_LICENSE("Proprietary");
MODULE_AUTHOR("Charles Lee @ charles@welcm.com");
MODULE_DESCRIPTION("EBC Interrupt Handler");
MODULE_SUPPORTED_DEVICE("EBC - AMCC Taihu board (PPC405EP)");
